#include "dispatcher.h"

#include "uploader.h"

Dispatcher::Dispatcher(Uploader *uploader, QObject *parent) :
    QObject(parent)
{
    m_client = rfbGetClient(8, 3, 4);

    m_client->canHandleNewFBSize = FALSE;
    m_client->MallocFrameBuffer = Dispatcher::MallocFrameBuffer;
    m_client->GotCopyRect = Dispatcher::GotCopyRect;
    m_client->GotFillRect = Dispatcher::GotFillRect;
    m_client->GotBitmap = Dispatcher::GotBitmap;
    m_client->GotJpeg = Dispatcher::GotJpeg;
    m_client->FinishedFrameBufferUpdate = Dispatcher::FinishedFrameBufferUpdate;
    m_client->GetPassword = Dispatcher::GetPassword;

    rfbClientSetClientData(m_client, const_cast<char*>("dispatcher"), this);
    // the pointer to the Uploader is stored in the rfbClient instead of a member of the class to avoid duplication
    rfbClientSetClientData(m_client, const_cast<char*>("uploader"), uploader);
}

Dispatcher::~Dispatcher()
{
    rfbClientCleanup(m_client);
}

rfbBool Dispatcher::MallocFrameBuffer(rfbClient *client)
{
    auto dispatcher = static_cast<Dispatcher *>(rfbClientGetClientData(client, const_cast<char*>("dispatcher")));
    emit dispatcher->resizeFramebuffer(QSize(client->width, client->height));
    return TRUE;
}

void Dispatcher::GotCopyRect(rfbClient *client, const int src_x, const int src_y, const int w, const int h, const int dest_x, const int dest_y)
{
    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(client, const_cast<char*>("uploader")));
    uploader->gotCopy(src_x, src_y, w, h, dest_x, dest_y);
}

void Dispatcher::GotFillRect(rfbClient *client, int x, int y, int w, int h, uint32_t colour)
{
    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(client, const_cast<char*>("uploader")));
    int red = (colour >> client->format.redShift) & client->format.redMax;
    int green = (colour >> client->format.greenShift) & client->format.greenMax;
    int blue = (colour >> client->format.blueShift) & client->format.blueMax;
    uploader->gotFill(x, y, w, h, red, green, blue);
}

void Dispatcher::GotBitmap(rfbClient *client, const uint8_t *buffer, int x, int y, int w, int h)
{
    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(client, const_cast<char*>("uploader")));
    uploader->gotBitmap(buffer, x, y, w, h);
}

rfbBool Dispatcher::GotJpeg(rfbClient *client, const uint8_t *buffer, int length, int x, int y, int w, int h)
{
    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(client, const_cast<char*>("uploader")));
    uploader->gotJpeg(buffer, length, x, y, w, h);
    return TRUE;
}

void Dispatcher::FinishedFrameBufferUpdate(rfbClient *client)
{
    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(client, const_cast<char*>("uploader")));
    uploader->finishedUpdate();
}

char *Dispatcher::GetPassword(rfbClient *client)
{
    char* password = (char*) std::malloc(sizeof(char) * 9);
    strcpy(password, "asdfasdf");
    return password;
}

bool Dispatcher::mouseEvent(const int x, const int y, const bool left, const bool middle, const bool right, const bool up, const bool down)
{
    if(m_client == NULL)
        return false;

    int buttonMask = 0;
    if(left) buttonMask |= rfbButton1Mask;
    if(right) buttonMask |= rfbButton2Mask;
    if(middle) buttonMask |= rfbButton3Mask;
    if(up) buttonMask |= rfbButton4Mask;
    if(down) buttonMask |= rfbButton5Mask;
    SendPointerEvent(m_client, x, y, buttonMask);
    return true;
}

int Dispatcher::bitsPerPixel()
{
    if(m_client == NULL)
        return 0;
    return m_client->format.bitsPerPixel;
}

void Dispatcher::open(const QString host, const int port, const int width, const int height)
{
    QByteArray tmp = host.toLocal8Bit();

    m_client->serverHost = new char[tmp.size() + 1];
    strcpy(m_client->serverHost, tmp.constData());
    m_client->serverPort = port;
    //m_client->appData.useRemoteCursor = TRUE;
    m_client->width = width;
    m_client->height = height;
    //m_client->appData.compressLevel = 0;
    //m_client->appData.qualityLevel = 9;
#ifdef __BIG_ENDIAN__
    m_client->format.bigEndian = TRUE;
#else
    m_client->format.bigEndian = FALSE;
#endif
    // The Raspberry Pi supports 8 bit paletted images and RGB422 too but only through extensions
    m_client->format.trueColour = TRUE;
    m_client->format.bitsPerPixel = 32;

    if(!rfbInitClient(m_client, NULL, NULL)) {
        emit error();
        return;
    }

    refresh();
    decode();
}
void Dispatcher::refresh()
{
    if(m_client == NULL)
        return;

    int message = 0;

    // Wait until we get something
    while(message == 0) {
        message = WaitForMessage(m_client, 1000);
        if(message < 0) {
            terminate();
            emit error();
            return;
        }
    }
}
void Dispatcher::decode()
{

    auto uploader = static_cast<Uploader *>(rfbClientGetClientData(m_client, const_cast<char*>("uploader")));
    uploader->startDecoding();

    if(!HandleRFBServerMessage(m_client)) {
        terminate();
        emit error();
        return;
    }
}

void Dispatcher::terminate()
{
    if(m_client == NULL)
        return;

    rfbClientCleanup(m_client);
}
